# Code for the Trie Data Structure.

import dill

class node():
    def __init__(self, char, parent, end=False):
        # Initialize all the variables used by a Trie Node.
        self.character = char  # The letter this node represents.
        self.parent = parent  # Pointer to the parent node, which is needed when removing keywords from the Trie.
        self.links = [None]*26  # There are 26 letters in the alphabet.
        self.ending = end  # If this node marks the end of a word.


class trie_class():
    def __init__(self):
        self.alpha = {}  # Alphabet and corresponding Index.
        for i in range(97,123):
            self.alpha[chr(i)] = i-97
        self.data = node('ROOT', 'is_root')  # The root node of the Trie.

    def add_keyword(self, word):
        # Adds any input string into the Trie.

        curr_index_value = self.data
        for letter in range(len(word)):
            letter_index = self.alpha[word[letter]]  # Converts character to an index from 0 to 25.
            if curr_index_value.links[letter_index] is None:
                if letter < len(word)-1:
                    # Adds node that is not end of word.
                    curr_index_value.links[letter_index] = node(word[letter], curr_index_value)
                else:
                    # Adds node that is end of word.
                    curr_index_value.links[letter_index] = node(word[letter], curr_index_value, True)
            # Check if word was already added.
            elif letter == len(word)-1 and curr_index_value.links[letter_index].ending:
                return f"The word '{word}' is already in the Trie"
            curr_index_value = curr_index_value.links[letter_index]
        return f"The word '{word}' was added to the Trie"

    def remove_keyword(self, word):
        # Remove any input string from the Trie, given it exists.

        curr_index_value = self.reach_node_required(self.data, word)
        # Return if word to remove does not exist.
        if type(curr_index_value) is str:
            return curr_index_value
        # If word never existed.
        if not curr_index_value.ending:
            return f"'{word}' was not in the Trie"
        # Set node ending to false if it has links to other nodes.
        links = 0
        for link in curr_index_value.links:
            links += 1 if link is not None else 0
        if links > 0:
            curr_index_value.ending = False
        # If node has no links, delete unused nodes.
        else:
            while True:
                if curr_index_value.character == 'ROOT':
                    break  # Break if ROOT is reached
                letter_index = self.alpha[curr_index_value.character]
                curr_index_value = curr_index_value.parent
                curr_index_value.links[letter_index] = None
                links = 0
                for link in curr_index_value.links:
                    links += 1 if link is not None else 0
                if links > 0:
                    break
        return f"The word '{word}' was removed from the Trie"

    def keyword_exists(self, word):
        # Checks if an input string is stored in the Trie.

        curr_index_value = self.reach_node_required(self.data, word)
        try:
            if curr_index_value.ending:
                return f"The word '{word}' exists in the Trie"
            return f"The word '{word}' does not exist in the Trie"
        except AttributeError:
            # Handles error when returned value is a string saying "This word does not exist"
            return curr_index_value

    def reach_node_required(self, node, prefix):
        # Finds the node for a certain word/prefix.

        for letter in range(len(prefix)):
            # For each letter, it further traverses down the Trie unless the needed letter does not exist.
            letter_index = self.alpha[prefix[letter]]
            if node.links[letter_index] is None:
                return f"The word {prefix} does not exist in the Trie"
            node = node.links[letter_index]
        return node

    def autocomplete(self, prefix):
        # Returns all strings in Trie that contain the prefix.

        def find_all_words(node, prefix):
            # Perform depth first search to get all words with required prefix.

            found_words = []
            for link in node.links:
                if link is not None:
                    if link.ending:
                        found_words.append(prefix + link.character)
                    found_words += find_all_words(link, prefix + link.character)
            return found_words

        words_with_prefix = []
        curr_index_value = self.reach_node_required(self.data, prefix)

        # Check if result was not error.
        if type(curr_index_value) is str:
            return words_with_prefix

        if curr_index_value.ending:
            words_with_prefix.append(prefix)
        words_with_prefix += find_all_words(curr_index_value, prefix)
        return words_with_prefix

    def print(self, node=False, indentation=0):
        # Returns a readable format of the Trie

        # If node not provided, use root node.
        if node is False:
            node = self.data

        output = [""]  # Array needs to be used so can be accessed in loop function

        def loop(node, indentation=0):
            for link in node.links:
                if link is not None:
                    output[0] += ' ' * indentation + link.character + "\n"
                    loop(link, indentation=indentation+2)

        loop(node)
        return output[0]





from setuptools import setup, find_packages

setup(
    name='globaltrie-server',
    version='1.0.7',
    author="dajkatal",
    author_email="<dajkatal@gmail.com>",
    description="CLI to host Trie Server",
    long_description_content_type="text/markdown",
    long_description="CLI to host a server with the Trie data structure, allowing multiple clients to concurrently interact with it.",
    url="https://github.com/dajkatal/Global-Trie",
    packages=find_packages(),
    install_requires=['dill'],
    entry_points = {
        "console_scripts": ['trieserver = globaltrieserver.server_cli:main']
    },
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ]
)


# Only change things unless you know what you are doing.

import os, dill, socket, threading, argparse
from requests import get
from .trie import trie_class


def main():
    # Set up the address of server.
    parser = argparse.ArgumentParser()
    parser.add_argument('-s', '--start', metavar='PORT', type=int, help="Mention a specific port to start the Trie Server")
    args = parser.parse_args()
    if args.start is None:
        # Make sure IP and PORT is filled out.

        print("You need to state a port on which to run the Trie Server.")
        exit()


    IP_to_connect = get('https://api.ipify.org').text  # IP clients need to connect to the server
    IP = '0.0.0.0'  # IP of Trie Server
    PORT = args.start  # Port of Trie Server
    ADDRESS = (IP, PORT)
    # Disconnect message to let clients disconnect.
    disconnect_message = 'disconnect'
    # Name of file storing trie.
    file_name = 'trie_dump'
    file_path = f"{os.path.expanduser('~')}/trieserver-data"
    file_location = f"{file_path}/{file_name}"

    # Check if directory where Trie is stored exists (Usually only when program is run for first time)
    if not os.path.isdir(file_path):
        os.makedirs(file_path)

    # Load Trie
    trie = None
    if os.path.isfile(file_location):
        with open(file_location, 'rb') as dill_file:
            trie = dill.load(dill_file)
        print("Loaded saved Trie")
    else:
        trie = trie_class()
        print("Trie was not saved... new Trie made")

    # Start server.
    try:
        print(f"[Starting] server is starting on {IP_to_connect}:{PORT}...")
        server = socket.socket()
        server.bind(ADDRESS)
    except OSError:
        # Happens if port is closed.
        print(f"Seems like Port {PORT} is either closed or already used. Try another PORT")
        exit()


    def save_trie():
        # If save command is run, the Trie is saved to a file.
        with open(file_location, "wb") as dill_file:
            dill.dump(trie, dill_file)


    def send_msg(response, connection):
        # Function to send message back to the client.

        msg_length_in_bytes = str(len(response)).encode('utf-8')
        connection.send(b' ' * (64 - len(msg_length_in_bytes)) + msg_length_in_bytes)
        connection.send(response.encode('utf-8'))


    queue = []
    needs_initiation = [True]


    def queue_handler():
        # Handles priority if multiple requests are made at once.

        # Instead of checking queue length, just try.
        try:
            # Check message.
            data = queue.pop()
            msg = data[0].lower()
            split_msg = msg.split(" ")
            # Make sure message contains only command word and input. (i.e ADD hello)
            if len(split_msg) > 2 or not msg.replace(" ", "").isalpha():
                send_msg("Wrong message format", data[1])
            else:
                first_word = split_msg[0]
                if first_word == 'save':
                    save_trie()
                    send_msg("The Trie was saved", data[1])
                elif first_word == 'add':
                    output = trie.add_keyword(split_msg[1])
                    send_msg(output, data[1])
                elif first_word == 'remove':
                    output = trie.remove_keyword(split_msg[1])
                    send_msg(output, data[1])
                elif first_word == 'exists':
                    output = trie.keyword_exists(split_msg[1])
                    send_msg(output, data[1])
                elif first_word == 'autocomplete':
                    output = trie.autocomplete(split_msg[1])
                    if len(output) == 0:
                        output = ['No Words']
                    send_msg(f"Words that start with '{split_msg[1]}' are: " + ', '.join(output), data[1])
                elif first_word == 'print':
                    output = trie.print()
                    if len(output) == 0:
                        output = 'Empty'
                    if len(output) == 0:
                        output = "No data in Trie"
                    send_msg('\n' + output, data[1])
                # Allow client to disconnect.
                elif msg == disconnect_message:
                    send_msg("disconnect", data[1])
                    data[1].close()
                else:
                    send_msg('This script has been tampered with. Make sure the command words are correct.', data[1])

            queue_handler()
        except IndexError:
            needs_initiation[0] = True


    def client_function(connection):
        while True:
            try:
                # Get message length.
                msg_length = connection.recv(64)
                msg_length = int(msg_length.decode('utf-8'))
                # Receive message.
                msg = connection.recv(msg_length).decode('utf-8')
                if msg == disconnect_message:
                    send_msg("disconnect", connection)
                    break
                queue.append([msg, connection])
                if needs_initiation[0]:
                    needs_initiation[0] = False
                    queue_handler()
            except:
                break
        connection.close()


    thread = threading.Thread(target=queue_handler)
    thread.start()
    server.listen(2)
    while True:
        client, address = server.accept()
        print(f"[Connected] connection made to {address[0]}:{address[1]}...")
        thread = threading.Thread(target=client_function, args=(client,))
        thread.start()
        print(f"[Clients] {threading.active_count()-1} active clients...")
